==> isolate_bloc-0.0.1.tar.gz: lib/src/common/isolate_bloc_wrapper.dart <==
import 'dart:async';

import 'package:flutter/foundation.dart';
import 'isolate_bloc.dart';
import 'service_events.dart';

import 'transition.dart';

/// {@template bloc_unhandled_error_exception}
/// Exception thrown in debug mode when an unhandled error occurs within a bloc.
/// {@endtemplate}
class BlocUnhandledErrorException implements Exception {
  /// The [bloc] in which the unhandled error occurred.
  final IsolateBlocWrapper bloc;

  /// The unhandled [error] object.
  final Object error;

  /// An optional [stackTrace] which accompanied the error.
  final StackTrace stackTrace;

  /// {@macro bloc_unhandled_error_exception}
  BlocUnhandledErrorException(this.bloc, this.error, [this.stackTrace]);

  @override
  String toString() {
    return 'Unhandled error $error occurred in bloc $bloc.\n'
        '${stackTrace ?? ''}';
  }
}

/// Signature for a mapper function which takes an [Event] as input
/// and outputs a [Stream] of [Transition] objects.
typedef TransitionFunction<Event, State> = Stream<Transition<Event, State>> Function(Event);

/// Signature for event receiver function which takes an [IsolateBlocTransitionEvent]
/// and send this event to the [IsolateBloc]
typedef EventReceiver = void Function(IsolateBlocTransitionEvent<Object> event);

/// {@template bloc}
/// This is implementation of
/// {@endtemplate}
class IsolateBlocWrapper<State> extends Stream<State> implements Sink<Object> {
  final _eventController = StreamController<Object>.broadcast();
  final _stateController = StreamController<State>.broadcast();

  /// Id of IsolateBloc. It's needed to find bloc in isolate.
  String _originBlocUuid;

  State _state;
  Object _event;
  final _unsendedEvents = <Object>[];
  final void Function(String uuid) _onBlocClose;
  StreamSubscription<Transition<Object, State>> _stateTransitionSubscription;
  StreamSubscription<Object> _eventReceiverSubscription;

  /// Returns stream with `event`
  Stream<Object> get eventStream => _eventController.stream;

  /// Callback which receive events and send them to the IsolateBloc
  final EventReceiver _eventReceiver;

  /// Returns the current [state] of the [bloc].
  State get state => _state;

  /// Returns whether the `Stream<State>` is a broadcast stream.
  @override
  bool get isBroadcast => _stateController.stream.isBroadcast;

  /// {@macro bloc}
  IsolateBlocWrapper(
    this._state,
    this._eventReceiver,
    this._onBlocClose,
  ) {
    _bindEventsListener();
  }

  /// Adds a subscription to the `Stream<State>`.
  /// Returns a [StreamSubscription] which handles events from
  /// the `Stream<State>` using the provided [onData], [onError] and [onDone]
  /// handlers.
  @override
  StreamSubscription<State> listen(
    void Function(State) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return _prepareStateStream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  Stream<State> get _prepareStateStream async* {
    yield state;
    yield* _stateController.stream;
  }

  /// Called whenever an [event] is [add]ed to the [bloc].
  /// A great spot to add logging/analytics at the individual [bloc] level.
  ///
  /// **Note: `super.onEvent` should always be called last.**
  /// ```dart
  /// @override
  /// void onEvent(Event event) {
  ///   // Custom onEvent logic goes here
  ///
  ///   // Always call super.onEvent with the current event
  ///   super.onEvent(event);
  /// }
  /// ```
  @mustCallSuper
  void onEvent(Object event) {
    // ignore: invalid_use_of_protected_member
    IsolateBloc.observer.onEvent(this, event);
  }

  /// Called whenever a [transition] occurs with the given [transition].
  /// A [transition] occurs when a new `event` is [add]ed and `state`
  /// from [_orginBlocStateStream] recived.
  /// [onTransition] is called before a [bloc]'s [state] has been updated.
  /// A great spot to add logging/analytics at the individual [bloc] level.
  ///
  /// **Note: `super.onTransition` should always be called last.**
  /// ```dart
  /// @override
  /// void onTransition(Transition<Event, State> transition) {
  ///   // Custom onTransition logic goes here
  ///
  ///   // Always call super.onTransition with the current transition
  ///   super.onTransition(transition);
  /// }
  /// ```
  @mustCallSuper
  void onTransition(Transition<Object, State> transition) {
    // ignore: invalid_use_of_protected_member
    IsolateBloc.observer.onTransition(this, transition);
  }

  /// Called whenever an [error] is thrown within onTransition in [IsolatedBloc].
  /// By default all [error]s will be ignored and [bloc] functionality will be
  /// unaffected.
  /// The [stackTrace] argument may be `null` if the [state] stream received
  /// an error without a [stackTrace].
  /// A great spot to handle errors at the individual [IsolateBlocWrapper] level.
  ///
  /// **Note: `super.onError` should always be called last.**
  /// ```dart
  /// @override
  /// void onError(Object error, StackTrace stackTrace) {
  ///   // Custom onError logic goes here
  ///
  ///   // Always call super.onError with the current error and stackTrace
  ///   super.onError(error, stackTrace);
  /// }
  /// ```
  @mustCallSuper
  void onError(Object error, [StackTrace stackTrace]) {
    // ignore: invalid_use_of_protected_member
    IsolateBloc.observer.onError(this, error, stackTrace);
    if (kDebugMode) {
      throw BlocUnhandledErrorException(this, error, stackTrace);
    }
  }

  /// Notifies the [bloc] of a new [event] which triggers onEvent in [IsolatedBloc].
  /// If [close] has already been called, any subsequent calls to [add] will
  /// be delegated to the [onError] method which can be overridden at the [bloc]
  /// as well as the [BlocDelegate] level.
  @override
  void add(Object event) {
    _event = event;
    try {
      onEvent(event);
      _eventController.add(event);
    } on dynamic catch (error, stackTrace) {
      onError(error, stackTrace);
    }
  }

  /// Closes the `event` and `state` `Streams`.
  /// This method should be called when a [bloc] is no longer needed.
  /// Once [close] is called, `events` that are [add]ed will not be
  /// processed and will result in an error being passed to [onError].
  /// In addition, if [close] is called while `events` are still being
  /// processed,
  /// the [bloc] will continue to process the pending `events` to completion.
  @override
  @mustCallSuper
  Future<void> close() async {
    _onBlocClose(_originBlocUuid);
    await _eventController.close();
    await _stateController.close();
    await _stateTransitionSubscription?.cancel();
    await _eventReceiverSubscription?.cancel();
  }

  /// Connect this wrapper to the origin IsolateBloc and start listening for state.
  void setUuid(String uuid) {
    assert(uuid != null);
    _originBlocUuid = uuid;
    while (_unsendedEvents.isNotEmpty) {
      _eventReceiver(IsolateBlocTransitionEvent<Object>(
        _originBlocUuid,
        _unsendedEvents.removeAt(0),
      ));
    }
  }

  /// Receive IsolateBloc's state and map them
  void stateReceiver(State nextState) {
    final transition = Transition(
      currentState: state,
      event: _event,
      nextState: nextState,
    );

    if (transition.nextState == state) return;
    try {
      onTransition(transition);
      _state = transition.nextState;
      _stateController.add(transition.nextState);
    } on dynamic catch (error, stackTrace) {
      onError(error, stackTrace);
    }
  }

  /// Start listening for new `events`
  void _bindEventsListener() {
    _eventReceiverSubscription = eventStream.listen((event) {
      if (_originBlocUuid != null) {
        _eventReceiver(IsolateBlocTransitionEvent<Object>(_originBlocUuid, event));
      } else {
        _unsendedEvents.add(event);
      }
    });
  }
}

==> isolate_bloc-0.1.0.tar.gz: lib/src/common/isolate_bloc_wrapper.dart <==
import 'dart:async';

import 'package:flutter/foundation.dart';
import 'isolate_bloc.dart';
import 'service_events.dart';

import 'transition.dart';

/// Signature for event receiver function which takes an [IsolateBlocTransitionEvent]
/// and send this event to the [IsolateBloc]
typedef EventReceiver = void Function(IsolateBlocTransitionEvent<Object> event);

/// Signature for function which takes [IsolateBloc]'s uuid and close this it
typedef BlocClosure = void Function(String uuid);

/// {@template bloc}
/// This is implementation of
/// {@endtemplate}
class IsolateBlocWrapper<State> extends Stream<State> implements Sink<Object> {
  final _eventController = StreamController<Object>.broadcast();
  final _stateController = StreamController<State>.broadcast();

  /// Id of IsolateBloc. It's needed to find bloc in isolate.
  String _originBlocUuid;

  State _state;
  bool _initStateProvided;
  final _unsentEvents = <Object>[];
  final BlocClosure _onBlocClose;
  StreamSubscription<Transition<Object, State>> _stateTransitionSubscription;
  StreamSubscription<Object> _eventReceiverSubscription;

  /// Returns stream with `event`
  Stream<Object> get eventStream => _eventController.stream;

  /// Callback which receive events and send them to the IsolateBloc
  final EventReceiver _eventReceiver;

  /// Returns the current [state] of the [bloc].
  State get state => _state;

  /// Returns whether the `Stream<State>` is a broadcast stream.
  @override
  bool get isBroadcast => _stateController.stream.isBroadcast;

  /// {@macro bloc}
  IsolateBlocWrapper(
    this._state,
    this._eventReceiver,
    this._onBlocClose,
  ) : _initStateProvided = true {
    _bindEventsListener();
  }

  IsolateBlocWrapper.noInitState(
    this._eventReceiver,
    this._onBlocClose,
  )   : _state = null,
        _initStateProvided = false {
    _bindEventsListener();
  }

  /// Adds a subscription to the `Stream<State>`.
  /// Returns a [StreamSubscription] which handles events from
  /// the `Stream<State>` using the provided [onData], [onError] and [onDone]
  /// handlers.
  @override
  StreamSubscription<State> listen(
    void Function(State) onData, {
    Function onError,
    void Function() onDone,
    bool cancelOnError,
  }) {
    return _prepareStateStream.listen(
      onData,
      onError: onError,
      onDone: onDone,
      cancelOnError: cancelOnError,
    );
  }

  Stream<State> get _prepareStateStream async* {
    if (_initStateProvided) {
      yield state;
    }
    yield* _stateController.stream;
  }

  /// Notifies the [bloc] of a new [event] which triggers onEvent in [IsolateBloc].
  @override
  void add(Object event) {
    _eventController.add(event);
  }

  /// Closes the `event` stream and request to close [IsolateBloc]
  @override
  @mustCallSuper
  Future<void> close() async {
    _onBlocClose(_originBlocUuid);
    await _eventController.close();
    await _stateController.close();
    await _stateTransitionSubscription?.cancel();
    await _eventReceiverSubscription?.cancel();
  }

  /// Connect this wrapper to the origin IsolateBloc and start listening for state.
  void setUuid(String uuid) {
    assert(uuid != null);
    _originBlocUuid = uuid;
    while (_unsentEvents.isNotEmpty) {
      _eventReceiver(IsolateBlocTransitionEvent<Object>(
        _originBlocUuid,
        _unsentEvents.removeAt(0),
      ));
    }
  }

  /// Receive IsolateBloc's state and map them
  void stateReceiver(State nextState) {
    _initStateProvided = true;
    _stateController.add(nextState);
  }

  /// Start listening for new `events`
  void _bindEventsListener() {
    _eventReceiverSubscription = eventStream.listen((event) {
      if (_originBlocUuid != null) {
        _eventReceiver(
            IsolateBlocTransitionEvent<Object>(_originBlocUuid, event));
      } else {
        _unsentEvents.add(event);
      }
    });
  }
}
